==> tsoutapl-1.2.tar.gz: tsoutapl-1.2/TypeSystem.hs <==
-- | Parser for types and contexts.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module TypeSystem
    ( Type(..)
    , Typing
    , Context
    , arrowize
    , baptize
    , showTypedTerm
    , showTypedComputation
    ) where

import Sidekicks
import Terms
import Calculus ( safefulleval )
import Maybe    ( fromMaybe )
import Char     ( toUpper )
import Text.ParserCombinators.Parsec.Error ( ParseError )


-- | Types.
data Type = TyBool
          | TyNat
          | Type `TyArrow` Type
          | TyCustom String
          | TyWrong String
          deriving (Eq)

instance Show Type where
    show TyBool                 = "Bool"
    show TyNat                  = "Nat"
    show (ty1 `TyArrow` ty2)    = "(" ++ show ty1 ++ " -> " ++ show ty2 ++ ")"
    show (TyCustom s)           = "" ++ s ++ ""
    show (TyWrong s)            = "<Wrong: " ++ s ++ ">"

type Typing = (String, Type)
type Context = [Typing]


-- | Given a context and a term, try to type the term according to the context.
typeOf :: Context -> Term -> Type
typeOf ctx term = case term of
    (TmTrue)        -> TyBool
    (TmFalse)       -> TyBool
    (TmIsZero t)    -> if (typeOf ctx t) == TyNat
                       then TyBool
                       else TyWrong "only numbers can be tested for zeroness"
    (TmZero)        -> TyNat
    (TmSucc t)      -> if (typeOf ctx t) == TyNat
                       then TyNat
                       else TyWrong "can only apply succ to numbers"
    (TmPred t)      -> if (typeOf ctx t) == TyNat
                       then TyNat
                       else TyWrong "can only apply pred to numbers"
    (TmVar x)       -> fromMaybe (baptize x) (lookup x ctx)
    (TmAbs x t)     -> if (lookup x ctx) == Nothing
                       then ((baptize x) `TyArrow` (typeOf ctx t))
                       else ((typeOf ctx (TmVar x)) `TyArrow` (typeOf ctx t))
    (t1 `TmApp` t2) -> let [ty1, ty2] = map (typeOf ctx) [t1, t2] in
                       appType ty1 ty2
    (TmIf t t1 t2)  -> let [ty, ty1, ty2] = map (typeOf ctx) [t, t1, t2] in
                       ifType ty ty1 ty2
    (TmLoop)        -> TyWrong "BOTTOM"
    _               -> TyWrong ("cannot type term " ++ show term)


{- Helper typers -}

-- | Make up a type name for the given variable.
baptize :: String -> Type
baptize var = TyCustom ("`" ++ (map toUpper var) ++ "`")


-- | Return /just/ the domain of an arrow.
dom :: Type -> Maybe Type
dom (ty1 `TyArrow` ty2) = Just ty1
dom _                   = Nothing


-- | Return /just/ the codomain of an arrow.
codom :: Type -> Maybe Type
codom (ty1 `TyArrow` ty2) = Just ty2
codom _                   = Nothing


-- Turn a list of types into an arrow by associating to the right.
-- (Something like foldr1 TyArrow types but with nifty TyWrong failures.)
arrowize :: [Type] -> Type
arrowize []      = TyWrong "null arrow"
arrowize [_]     = TyWrong "poor arrow"
arrowize [t1,t2] = t1 `TyArrow` t2
arrowize (t:ts)  = t `TyArrow` (arrowize ts)


-- | Try to type two given types as an application.
appType :: Type -> Type -> Type
appType ty1 ty2 = case (dom ty1) of
    Nothing      -> TyWrong (show ty1 ++ " is not an arrow")
    Just ty1dom  -> if ty1dom == ty2
                    then fromMaybe
                         (TyWrong ("cannot infer the codomain of " ++ show ty1))
                         (codom ty1)
                    else TyWrong ("arrow " ++ show ty1dom ++ " should have domain " ++ show ty2)

-- | Try to type three given types as a branching.
ifType :: Type -> Type -> Type -> Type
ifType ty ty1 ty2
    | ty /= TyBool = TyWrong "predicate is not boolean"
    | ty1 == ty2   = ty1
    | otherwise    = TyWrong ("either " ++ show ty1 ++ " or " ++ show ty2)


-- | Pretty-print a term and its type in a given context.
showTypedTerm :: Context -> Term -> String
showTypedTerm ctx t = show t ++ " :: " ++ show (typeOf ctx t)


-- | Pretty-print a series of terms as a typed computation.
showTypedComputation :: Context -> Computation -> String
showTypedComputation ctx ts = joinWith "\n ==> " (map (showTypedTerm ctx) ts) ++ "\n"

==> tsoutapl-1.3.tar.gz: tsoutapl-1.3/TypeSystem.hs <==
-- | Parser for types and contexts.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module TypeSystem
    ( Type(..)
    , Typing
    , Context
    , arrowize
    , baptize
    , showTypedTerm
    , showTypedComputation
    ) where

import Sidekicks
import Terms
import Calculus ( safefulleval )
import Maybe    ( fromMaybe )
import Char     ( toUpper )
import Text.ParserCombinators.Parsec.Error ( ParseError )


-- | Types for this type system.
data Type = TyBool
          | TyNat
          | Type `TyArrow` Type
          | TyCustom String
          | TyWrong String
          deriving (Eq)

instance Show Type where
    show TyBool                 = "Bool"
    show TyNat                  = "Nat"
    show (ty1 `TyArrow` ty2)    = "(" ++ show ty1 ++ " -> " ++ show ty2 ++ ")"
    show (TyCustom s)           = "" ++ s ++ ""
    show (TyWrong s)            = "<Wrong: " ++ s ++ ">"

-- | A Typing of a variable (name) to a Type.
type Typing = (String, Type)

-- | A series of Typings
type Context = [Typing]


-- | Given a context and a term, try to type the term according to the context.
typeOf :: Context -> Term -> Type
typeOf ctx term = case term of
    (TmTrue)        -> TyBool
    (TmFalse)       -> TyBool
    (TmIsZero t)    -> if (typeOf ctx t) == TyNat
                       then TyBool
                       else TyWrong "only numbers can be tested for zeroness"
    (TmZero)        -> TyNat
    (TmSucc t)      -> if (typeOf ctx t) == TyNat
                       then TyNat
                       else TyWrong "can only apply succ to numbers"
    (TmPred t)      -> if (typeOf ctx t) == TyNat
                       then TyNat
                       else TyWrong "can only apply pred to numbers"
    (TmVar x)       -> fromMaybe (baptize x) (lookup x ctx)
    (TmAbs x t)     -> if (lookup x ctx) == Nothing
                       then ((baptize x) `TyArrow` (typeOf ctx t))
                       else ((typeOf ctx (TmVar x)) `TyArrow` (typeOf ctx t))
    (t1 `TmApp` t2) -> let [ty1, ty2] = map (typeOf ctx) [t1, t2] in
                       appType ty1 ty2
    (TmIf t t1 t2)  -> let [ty, ty1, ty2] = map (typeOf ctx) [t, t1, t2] in
                       ifType ty ty1 ty2
    (TmLoop)        -> TyWrong "BOTTOM"
    _               -> TyWrong ("cannot type term " ++ show term)


{- Helper typers -}

-- | Make up a type name for the given variable.
baptize :: String -> Type
baptize var = TyCustom ("`" ++ (map toUpper var) ++ "`")


-- | Return /just/ the domain of an arrow.
dom :: Type -> Maybe Type
dom (ty1 `TyArrow` ty2) = Just ty1
dom _                   = Nothing


-- | Return /just/ the codomain of an arrow.
codom :: Type -> Maybe Type
codom (ty1 `TyArrow` ty2) = Just ty2
codom _                   = Nothing


-- | Turn a list of types into an arrow by associating to the right.
-- (Something like foldr1 TyArrow types but with nifty TyWrong failures.)
arrowize :: [Type] -> Type
arrowize []      = TyWrong "null arrow"
arrowize [_]     = TyWrong "poor arrow"
arrowize [t1,t2] = t1 `TyArrow` t2
arrowize (t:ts)  = t `TyArrow` (arrowize ts)


-- | Try to type two given types as an application.
appType :: Type -> Type -> Type
appType ty1 ty2 = case (dom ty1) of
    Nothing      -> TyWrong (show ty1 ++ " is not an arrow")
    Just ty1dom  -> if ty1dom == ty2
                    then fromMaybe
                         (TyWrong ("cannot infer the codomain of " ++ show ty1))
                         (codom ty1)
                    else TyWrong ("arrow " ++ show ty1dom ++ " should have domain " ++ show ty2)


-- | Try to type three given types as a branching.
ifType :: Type -> Type -> Type -> Type
ifType ty ty1 ty2
    | ty /= TyBool = TyWrong "predicate is not boolean"
    | ty1 == ty2   = ty1
    | otherwise    = TyWrong ("either " ++ show ty1 ++ " or " ++ show ty2)


-- | Pretty-print a term and its type in a given context.
showTypedTerm :: Context -> Term -> String
showTypedTerm ctx t = show t ++ " :: " ++ show (typeOf ctx t)


-- | Pretty-print a series of terms as a typed computation.
showTypedComputation :: Context -> Computation -> String
showTypedComputation ctx ts = joinWith "\n ==> " (map (showTypedTerm ctx) ts) ++ "\n"
